from __future__ import annotations

from contextlib import contextmanager
from ctypes import byref, c_void_p
from enum import Flag, IntEnum, auto
import io
import sys
from types import TracebackType
from typing import Any, Type

from pydantic import BaseModel, Field

from ._api import LIBNUX
from .errors import into_exception, is_err


class Resource(Flag):
    """Profiler target resource to be recorded."""

    CPU = auto()
    NPU = auto()
    ALL = CPU | NPU


class RecordFormat(IntEnum):
    """Profiler format to record profile data."""

    ChromeTrace = 0


class ChromeTraceConfig(BaseModel):
    """ChromeTrace specific config.

    Attributes:
        file: file descriptor to write profile data. By default, sys.stdout.
    """

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = True

        # Custom encoder for File like type.
        # See https://pydantic-docs.helpmanual.io/usage/exporting_models/#json_encoders
        # Here we uses C module "_io" as actual type which pydantic will use is "_io._IOBase".
        import _io

        json_encoders = {_io._IOBase: lambda f: f.fileno()}

    # Provide default value as default_factory because stdout itself is not pickle-able.
    file: io.IOBase = Field(default_factory=lambda: sys.stdout)


# Format specific configs
configs = {RecordFormat.ChromeTrace: ChromeTraceConfig}


class profile:
    """Profiler context manager.

    Examples:
        >>> from furiosa.runtime.profiler import RecordFormat
        >>> with open("profile.json", "w") as f:
        >>>     with profile(format=RecordFormat.ChromeTrace, file=f) as profiler:
        >>>         # Profiler enabled from here
        >>>         with profiler.record("Inference"):
        >>>             ... # Profiler recorded with span named 'Inference'
    """

    def __init__(
        self,
        resource: Resource = Resource.ALL,
        format: RecordFormat = RecordFormat.ChromeTrace,
        **config: Any,
    ):
        """Create profiler context with specified arguments.

        Args:
            resource (Resource): Target resource to be profiled. e.g. CPU or NPU.
            format (RecordFormat): Profiler format. e.g. ChromeTrace.
            **config: Format specific config. You need to pass valid arguments for the format.

        Raises:
           pydantic.error_wrappers.ValidationError: Raise when config validation failed.
        """
        formats = ", ".join(map(str, RecordFormat))
        assert format in configs, f"Invalid format: {format}. Choose one of {formats}"

        self.resource = resource
        self.format = format
        self.config = configs[format](**config).json()

    def __enter__(
        self,
    ):
        """Enter profiler context.

        This enables profiler until this context exit.

        Returns:
           profile: Profiler context manager.

        Raises:
            errors.InvalidYamlException: Raise when config (YAML compatible expected) is invalid.
        """

        self.profiler = c_void_p(None)
        err = LIBNUX.profiler_enable(
            self.resource.value, self.format.value, self.config.encode(), byref(self.profiler)
        )

        if is_err(err):
            raise into_exception(err)

        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ):
        """Exit profiler context.

        This disables the created profiler.
        """

        LIBNUX.profiler_disable(self.profiler)

    @contextmanager
    def record(
        self,
        name: str = "",
        warm_up=False,
    ):
        """Create profiler span with specified name.

        Args:
            name (str): Profiler record span name.
            warm_up (bool): If true, do not record profiler result, and just warm up.
        """

        span = LIBNUX.profiler_record_start(name.encode(), warm_up)
        yield
        LIBNUX.profiler_record_end(span)



import random
from typing import List
import unittest

import mnist
import numpy as np

from furiosa.runtime import errors, session
from tests.test_base import MNIST_ONNX, NAMED_TENSORS_ONNX, SessionTester, ensure_test_device

NPU_DEVICE_READY = ensure_test_device()


class TestSession(unittest.TestCase):
    mnist_images = mnist.train_images().reshape((60000, 1, 28, 28)).astype(np.float32)

    @classmethod
    def setUpClass(cls):
        cls.tester = SessionTester(MNIST_ONNX)

    @classmethod
    def tearDownClass(cls):
        cls.tester.close()

    def test_run(self):
        sess = self.tester.session

        self.assertEqual(
            sess.summary(),
            """\
Inputs:
{0: TensorDesc(name="Input3", shape=(1, 1, 28, 28), dtype=FLOAT32, format=NCHW, size=3136, len=784)}
Outputs:
{0: TensorDesc(name="Plus214_Output_0", shape=(1, 10), dtype=FLOAT32, format=??, size=40, len=10)}""",
        )

        items = 50
        for _ in range(0, items):
            idx = random.randrange(0, 9999, 1)
            ndarray_value = self.mnist_images[idx : idx + 1]
            self.assertEqual(ndarray_value.shape, sess.input(0).shape)
            self.assertEqual(ndarray_value.dtype, sess.input(0).dtype.numpy_dtype)

            result1 = sess.run(ndarray_value)
            result2 = sess.run([ndarray_value])
            self.assertTrue(np.array_equal(result1.numpy(), result2.numpy()))

            result3 = sess.run_with(["Plus214_Output_0"], {"Input3": ndarray_value})
            self.assertTrue(np.array_equal(result1.numpy(), result3.numpy()))

    def test_run_invalid_input(self):
        sess = self.tester.session
        self.assertRaises(
            errors.InvalidInput, lambda: sess.run(np.zeros((1, 28, 28, 1), dtype=np.float16))
        )
        self.assertRaises(
            errors.InvalidInput,
            lambda: sess.run(np.zeros((2, 28, 28, 2), dtype=sess.input(0).dtype.numpy_dtype)),
        )
        self.assertTrue(sess.run([np.zeros((1, 1, 28, 28), np.float32)]))
        self.assertRaises(
            errors.InvalidInput,
            lambda: sess.run(
                [np.zeros((1, 28, 28, 1), np.uint8), np.zeros((1, 28, 28, 1), np.uint8)]
            ),
        )
        self.assertRaises(
            errors.InvalidInput,
            lambda: sess.run_with(None, {"Input3": np.zeros((1, 28, 28, 1), np.uint8)}),
        )


@unittest.skipIf(not NPU_DEVICE_READY, "No NPU device")
class TestDeviceBusy(unittest.TestCase):
    def test_device_busy(self):
        with session.create(MNIST_ONNX) as _:
            self.assertRaises(errors.DeviceBusy, lambda: session.create(MNIST_ONNX))
            pass


class TestSessionClosed(unittest.TestCase):
    def test_session_closed(self):
        sess = session.create(MNIST_ONNX)
        sess.close()
        self.assertRaises(errors.SessionClosed, lambda: sess.inputs())
        self.assertRaises(errors.SessionClosed, lambda: sess.input_num())
        self.assertRaises(errors.SessionClosed, lambda: sess.outputs())
        self.assertRaises(errors.SessionClosed, lambda: sess.output_num())
        self.assertRaises(errors.SessionClosed, lambda: sess.summary())
        self.assertRaises(errors.SessionClosed, lambda: sess.run(None))

    def test_async_session_closed(self):
        submitter, queue = session.create_async(MNIST_ONNX)
        submitter.close()
        self.assertRaises(errors.SessionClosed, lambda: submitter.inputs())
        self.assertRaises(errors.SessionClosed, lambda: submitter.input_num())
        self.assertRaises(errors.SessionClosed, lambda: submitter.outputs())
        self.assertRaises(errors.SessionClosed, lambda: submitter.output_num())
        self.assertRaises(errors.SessionClosed, lambda: submitter.summary())
        queue.close()

    def test_queue_when_session_closed(self):
        submitter, queue = session.create_async(MNIST_ONNX)
        queue.close()
        self.assertRaises(errors.SessionClosed, lambda: queue.recv())
        submitter.close()


class TestSessionOptions(unittest.TestCase):
    def test_create(self):
        with session.create(
            MNIST_ONNX, worker_num=1, compile_config={"split_after_lower": True}
        ) as _:
            pass

    def test_batch(self):
        with session.create(
            MNIST_ONNX,
            batch_size=1,
        ) as sess:
            self.assertEqual(sess.input(0).shape, (1, 1, 28, 28))
            pass
        with session.create(
            MNIST_ONNX,
            batch_size=4,
        ) as sess:
            self.assertEqual(sess.input(0).shape, (4, 1, 28, 28))
            pass


class TestSessionWithNames(unittest.TestCase):
    def assert_named_tensors(self, sess, input_indices: List[int], output_indices: List[int]):
        from furiosa.runtime import tensor

        inputs = [sess.input(idx) for idx in input_indices]
        output_names = [sess.output(idx).name for idx in output_indices]
        outputs = sess.run_with(
            output_names,
            {
                inputs[0].name: tensor.rand(inputs[0]),
                inputs[1].name: tensor.rand(inputs[1]),
                inputs[2].name: tensor.rand(inputs[2]),
            },
        )

        for idx in range(0, len(outputs)):
            self.assertEqual(sess.output(output_indices[idx]).shape, outputs[idx].shape)
            self.assertEqual(sess.output(output_indices[idx]).dtype, outputs[idx].dtype)

    def test_run_with(self):
        with session.create(NAMED_TENSORS_ONNX) as sess:
            expected_inputs = ["input.1", "input.3", "input"]
            self.assertEqual(sess.input_num, 3)
            for idx in range(0, sess.input_num):
                self.assertEqual(sess.input(idx).name, expected_inputs[idx])

            self.assertEqual(sess.output_num, 3)
            expected_outputs = ["18_dequantized", "19_dequantized", "15_dequantized"]
            for idx in range(0, sess.output_num):
                self.assertEqual(sess.output(idx).name, expected_outputs[idx])

            self.assert_named_tensors(sess, [0, 1, 2], [0, 1, 2])
            self.assert_named_tensors(sess, [2, 0, 1], [0, 1, 2])
            self.assert_named_tensors(sess, [1, 2, 0], [0, 1, 2])

            self.assert_named_tensors(sess, [2, 1, 0], [0, 2, 1])
            self.assert_named_tensors(sess, [2, 1, 0], [1, 2, 0])
            self.assert_named_tensors(sess, [2, 1, 0], [2, 0, 1])
            self.assert_named_tensors(sess, [2, 1, 0], [2, 1, 0])

    def test_run_with_invalid_names(self):
        from furiosa.runtime import tensor

        with session.create(MNIST_ONNX) as sess:
            self.assertEqual(sess.input(0).name, "Input3")
            self.assertEqual(sess.output(0).name, "Plus214_Output_0")
            sess.run_with(["Plus214_Output_0"], {"Input3": tensor.zeros(sess.input(0))})

            # Wrong output
            self.assertRaises(
                errors.InvalidInput,
                lambda: sess.run_with(["WrongOutput"], {"Input3": tensor.zeros(sess.input(0))}),
            )
            # Wrong input
            self.assertRaises(
                errors.InvalidInput,
                lambda: sess.run_with(
                    ["Plus214_Output_0"], {"WrongInput3": tensor.zeros(sess.input(0))}
                ),
            )


class TestCompilerConfig(unittest.TestCase):
    def test_non_existence_config(self):
        non_existence_config = {"non_existence_config": 1}
        with session.create(MNIST_ONNX, compile_config=non_existence_config) as _:
            pass

    def test_invalid_config(self):  # put an existing config but use a different type
        invalid_config = {"remove_lower": 1}  # the correct value type is boolean
        self.assertRaises(
            errors.InvalidCompilerConfig,
            lambda: session.create(MNIST_ONNX, compile_config=invalid_config),
        )


if __name__ == '__main__':
    unittest.main()



import json
import tempfile
import unittest

import numpy as np

from furiosa.runtime import session
from furiosa.runtime.profiler import profile
from tests.test_base import MNIST_TFLITE_QUANTIZED


class TestProfiler(unittest.TestCase):
    def test_profile(self):
        with tempfile.TemporaryFile() as f:
            # Record profile data into temporary file
            with profile(file=f) as profiler:
                with session.create(MNIST_TFLITE_QUANTIZED) as sess:
                    input_meta = sess.inputs()[0]

                    input = np.random.randint(0, 127, input_meta.shape, dtype=np.uint8)

                    with profiler.record("Run"):
                        sess.run(input)

            f.seek(0)

            records = json.loads(f.read())
            self.assertTrue(len([record for record in records if record["name"] == "Run"]) == 2)



"""Utility module"""
import os
import sys
from typing import Optional

from furiosa.common.native import find_native_lib_path
from furiosa.runtime import __version__, runtime_version


def list_to_dict(items: list) -> dict:
    """Transform a list to a dict with keys derived from indexes"""
    dict_values = {}
    for idx, value in enumerate(items):
        dict_values[idx] = value

    return dict_values


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def dump_info(log_path: Optional[str] = None):
    width = 100  # default
    try:
        width, _ = os.get_terminal_size()
    except:
        pass

    eprint("=" * width)
    eprint("Information Dump")
    eprint("=" * width)
    eprint("- Python version: " + sys.version.replace('\n', ' '))
    eprint(f"- furiosa-libnux path: {find_native_lib_path('nux')}")
    eprint(f"- furiosa-libnux version: {runtime_version()}")
    eprint(f"- furiosa-compiler version: {runtime_version()}")
    eprint(f"- furiosa-runtime version: {__version__}")
    if log_path:
        eprint(
            f"\nPlease check the compiler log at {log_path}.\n"
            f"If you have a problem, please report the log file to https://furiosa-ai.atlassian.net/servicedesk/customer/portals\n"
            f"with the information dumped above."
        )
    eprint("=" * width)



"""Nux Exception and Error"""
import ctypes
from enum import IntEnum
import typing
from typing import Optional

from furiosa.common.error import FuriosaError, is_err, is_ok


class NativeError(IntEnum):
    """Python object correspondnig to nux_error_t in Nux C API"""

    SUCCESS = 0
    NUX_CREATION_FAILED = 1
    MODEL_DEPLOY_FAILED = 2
    MODEL_EXECUTION_FAILED = 3
    INVALID_INPUT_INDEX = 4
    INVALID_OUTPUT_INDEX = 5
    INVALID_BUFFER = 6
    INVALID_INPUTS = 7
    INVALID_OUTPUTS = 8
    GET_TASK_FAILED = 9
    DUMP_PROFILE_FAILED = 10
    QUEUE_WAIT_TIMEOUT = 11
    QUEUE_NO_DATA = 12
    INCOMPATIBLE_MODEL = 13
    COMPILATION_FAILED = 14
    INTERNAL_ERROR = 15
    INVALID_YAML = 16
    INCOMPATIBLE_API_CLIENT_ERROR = 17
    API_CLIENT_INIT_FAILED = 18
    NO_API_KEY = 19
    NULL_POINTER_EXCEPTION = 20
    INVALID_SESSION_OPTIONS = 21
    SESSION_TERMINATED = 22
    DEVICE_BUSY = 23
    TENSOR_NAME_NOT_FOUND = 24
    UNSUPPORTED_FEATURE = 25
    INVALID_COMPILER_CONFIG = 30


class NativeException(FuriosaError):
    """general exception caused by Nuxpy"""

    _native_err: Optional[NativeError]

    def __init__(self, message: str, native_err: NativeError = None):
        self._native_err = native_err
        super().__init__(message)

    def native_error(self) -> Optional[NativeError]:
        """Return a native error if this exception comes from C native extension"""
        return self._native_err

    def __repr__(self):
        if self._native_err is None:
            return self._message

        return f'{self._message} (native error code: {self._native_err})'

    def __str__(self):
        return self.__repr__()


class IncompatibleModel(NativeException):
    """When Renegade compiler cannot recognize a given model image binary"""

    def __init__(self):
        super().__init__("model binary is not compatible", NativeError.INCOMPATIBLE_MODEL)


class CompilationFailed(NativeException):
    """when Nux fails to compile a given model image to NPU model binary"""

    def __init__(self):
        super().__init__(
            "fail to compile a given model to NPU binary", NativeError.COMPILATION_FAILED
        )


class InternalError(NativeException):
    """internal error or no corresponding error in Python binding"""

    def __init__(self, cause='unknown'):
        super().__init__("{}".format(cause), NativeError.INTERNAL_ERROR)


class UnsupportedTensorType(NativeException):
    """Unsupported tensor type"""

    def __init__(self):
        super().__init__(
            "numpy.ndarray, TensorArray are only supported", NativeError.INVALID_INPUTS
        )


class UnsupportedDataType(NativeException):
    """Unsupported tensor data type"""

    def __init__(self, dtype):
        super().__init__("unknown data type: {}".format(dtype))


class IncompatibleApiClientError(NativeException):
    """When both API client and server are incompatible"""

    def __init__(self):
        super().__init__(
            "incompatible client with Furiosa API server", NativeError.INCOMPATIBLE_API_CLIENT_ERROR
        )


class InvalidYamlException(NativeException):
    """When Renegade compiler cannot recognize a given model image binary"""

    def __init__(self):
        super().__init__("compiler config is not valid YAML", NativeError.INVALID_YAML)


class ApiClientInitFailed(NativeException):
    """when api client fails to initialize due to api keys or others"""

    def __init__(self):
        super().__init__("fail to initialize API client", NativeError.API_CLIENT_INIT_FAILED)


class NoApiKeyException(NativeException):
    """when api client fails to initialize due to api keys or others"""

    def __init__(self):
        super().__init__("no API keys. Please check your API keys.", NativeError.NO_API_KEY)


class InvalidSessionOption(NativeException):
    """when api client fails to initialize due to api keys or others"""

    def __init__(self):
        super().__init__(
            "invalid options passed to session.create() or create_async()",
            NativeError.INVALID_SESSION_OPTIONS,
        )


class QueueWaitTimeout(NativeException):
    """Timed out in Completion queue"""

    def __init__(self):
        super().__init__("queue waiting timed out", NativeError.QUEUE_WAIT_TIMEOUT)


class SessionTerminated(NativeException):
    """Session is already terminated"""

    def __init__(self):
        super().__init__("Session or AsyncSession terminated", NativeError.SESSION_TERMINATED)


class DeviceBusy(NativeException):
    """The device is already occupied"""

    def __init__(self):
        super().__init__("NPU device busy", NativeError.DEVICE_BUSY)


class InvalidInput(FuriosaError):
    """When input tensors are invalid with any reason"""

    def __init__(self, message: str = "Invalid input tensors"):
        super().__init__(message)


class TensorNameNotFound(NativeException):
    """When a given tensor name is not found in this model"""

    def __init__(self):
        super().__init__("Tensor name not found", NativeError.TENSOR_NAME_NOT_FOUND)


class UnsupportedFeature(NativeException):
    """Feature is not supported"""

    def __init__(self):
        super().__init__("Unsupported feature", NativeError.UNSUPPORTED_FEATURE)


class InvalidCompilerConfig(NativeException):
    """Compiler config is invalid"""

    def __init__(self):
        super().__init__("Invalid compiler config", NativeError.INVALID_COMPILER_CONFIG)


class SessionClosed(FuriosaError):
    """Session is already terminated"""

    def __init__(self):
        super().__init__("Session or AsyncSession is already closed. Please create a new session.")


_errors_to_exceptions = {
    NativeError.INCOMPATIBLE_MODEL: IncompatibleModel(),
    NativeError.COMPILATION_FAILED: CompilationFailed(),
    NativeError.INTERNAL_ERROR: InternalError(),
    NativeError.INCOMPATIBLE_API_CLIENT_ERROR: IncompatibleApiClientError(),
    NativeError.INVALID_YAML: InvalidYamlException(),
    NativeError.API_CLIENT_INIT_FAILED: ApiClientInitFailed(),
    NativeError.NO_API_KEY: NoApiKeyException(),
    NativeError.INVALID_SESSION_OPTIONS: InvalidSessionOption(),
    NativeError.QUEUE_WAIT_TIMEOUT: QueueWaitTimeout(),
    NativeError.SESSION_TERMINATED: SessionTerminated(),
    NativeError.DEVICE_BUSY: DeviceBusy(),
    NativeError.TENSOR_NAME_NOT_FOUND: TensorNameNotFound(),
    NativeError.UNSUPPORTED_FEATURE: UnsupportedFeature(),
    NativeError.INVALID_COMPILER_CONFIG: InvalidCompilerConfig(),
}


def into_exception(err: typing.Union[ctypes.c_int, int]) -> NativeException:
    """
    Convert nux_error_t type in Nux C API to NuxException

    Arguments:
        err (NativeError) NativeError converted from ``nux_error_t`` enum in C

    Returns:
        NuxException
    """
    # FIXME (@hyunsik): C APIs defined in ctypes returns c_int, or int value.
    #   There's no way to make the behavior deterministic.
    err = err.value if isinstance(err, ctypes.c_int) else err

    if err == NativeError.SUCCESS:
        return RuntimeError(msg='NuxErr.SUCCESS cannot be NuxException')

    if err in _errors_to_exceptions:
        return _errors_to_exceptions[err]

    return InternalError()



import random
import unittest

import mnist
import numpy as np

from furiosa.runtime import errors, session
from tests.test_base import MNIST_ONNX, AsyncSessionTester, ensure_test_device

NPU_DEVICE_READY = ensure_test_device()


class TestAsyncSession(unittest.TestCase):
    mnist_images = mnist.train_images().reshape((60000, 1, 28, 28)).astype(np.float32)

    @classmethod
    def setUpClass(cls):
        cls.tester = AsyncSessionTester(MNIST_ONNX)

    @classmethod
    def tearDownClass(cls):
        cls.tester.close()

    def test_run_async(self):
        async_sess = self.tester.session

        self.assertEqual(
            async_sess.summary(),
            """\
Inputs:
{0: TensorDesc(name="Input3", shape=(1, 1, 28, 28), dtype=FLOAT32, format=NCHW, size=3136, len=784)}
Outputs:
{0: TensorDesc(name="Plus214_Output_0", shape=(1, 10), dtype=FLOAT32, format=??, size=40, len=10)}""",
        )

        items = 50
        for i in range(0, items):
            idx = random.randrange(0, 9999, 1)
            ndarray_value = self.mnist_images[idx : idx + 1]

            self.assertEqual(ndarray_value.shape, async_sess.input(0).shape)
            self.assertEqual(ndarray_value.dtype, async_sess.input(0).dtype.numpy_dtype)

            async_sess.submit([ndarray_value], i)

        keys = set()
        for key, _ in self.tester.queue:
            keys.add(key)

            if len(keys) == items:
                break

        self.assertEqual(set(range(0, items)), keys)


class TestAsyncSessionExceptions(unittest.TestCase):
    def test_create(self):
        sess = None
        queue = None
        try:
            sess, queue = session.create_async(
                MNIST_ONNX,
                worker_num=1,
                input_queue_size=1,
                output_queue_size=1,
                compile_config={"split_after_lower": True},
            )
        finally:
            if queue:
                queue.close()
            if sess:
                sess.close()

    def test_next_with_timeout(self):
        nux_sess = None
        nux_queue = None
        try:
            nux_sess, nux_queue = session.create_async(model=MNIST_ONNX)
            self.assertRaises(errors.QueueWaitTimeout, lambda: nux_queue.recv(timeout=100))
            nux_sess.close()
            self.assertRaises(errors.SessionTerminated, lambda: nux_queue.recv())
            self.assertRaises(errors.SessionTerminated, lambda: nux_queue.recv(timeout=100))
        except:
            if nux_sess:
                nux_sess.close()
            if nux_queue:
                nux_queue.close()


@unittest.skipIf(not NPU_DEVICE_READY, "No NPU device")
class TestDeviceBusy(unittest.TestCase):
    def test_device_busy(self):
        sess = None
        queue = None
        try:
            sess, queue = session.create_async(MNIST_ONNX)
            self.assertRaises(errors.DeviceBusy, lambda: session.create_async(MNIST_ONNX))
        finally:
            if sess:
                sess.close()
            if queue:
                queue.close()


if __name__ == '__main__':
    unittest.main()



from datetime import datetime
import logging
from pathlib import Path
import random
import string
from typing import Union

from . import envs
from ._util import eprint

LOG = logging.getLogger(__name__)


def _read_file(path: Union[str, Path]):
    with open(path, 'rb') as file:
        contents = file.read()
        return contents


def _model_image(model: Union[bytes, str, Path]) -> bytes:
    if isinstance(model, bytes):
        model_image = model
    elif isinstance(model, (str, Path)):
        model_image = Path(model).read_bytes()
    else:
        raise TypeError("'model' must be str or bytes, but it was " + repr(type(model)))

    return model_image


def _generate_suffix(length: int) -> str:
    return ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(length))


def _generate_unique_log_filename(log_dir: Path) -> Path:
    log_time = datetime.now().strftime(f"%Y%m%d%H%M%S")
    while True:
        log_filename = f"compile-{log_time}-{_generate_suffix(6)}.log"
        path = Path(f"{log_dir}/{log_filename}")
        try:
            path.touch(mode=0o644, exist_ok=False)
        except FileExistsError:
            continue

        return path


def generate_compiler_log_path() -> Path:
    """Generate a log path for compilation log"""
    if envs.is_compile_log_enabled():
        compiler_log_path = Path(envs.log_dir())
        compiler_log_path.mkdir(mode=0o755, parents=True, exist_ok=True)
        compiler_log_path = _generate_unique_log_filename(compiler_log_path)
        eprint(f"Saving the compilation log into {compiler_log_path}")
        return compiler_log_path
    else:
        return None



from __future__ import annotations

import os
from pathlib import Path
import warnings

from . import consts


def is_compile_log_enabled() -> bool:
    """
    Return True or False whether the compile log is enabled or not.

    Returns
        True if the compile log is enabled, or False.
    """
    return not os.environ.get('FURIOSA_TEST_IS_RUNNING')


def xdg_state_home() -> str:
    """
    Return XDG_STATE_HOME which is the base directory of furiosa logs, history, and other states

    Returns:
        Furiosa home directory
    """
    return os.environ.get(consts.ENV_XDG_STATE_HOME, f"{Path.home()}/.local/state/furiosa")


def log_dir() -> str:
    """
    Return FURIOSA_LOG_DIR where the logs are stored.

    Returns:
        The log directory of furiosa sdk
    """
    return os.environ.get(consts.ENV_FURIOSA_LOG_DIR, f"{xdg_state_home()}/logs")


def current_npu_device() -> str:
    """
    Return the current npu device name

    Returns:
        NPU device name
    """
    return os.environ.get(consts.ENV_NPU_DEVNAME, consts.DEFAULT_DEVNAME)


def profiler_output() -> None | str:
    """
    Return FURIOSA_PROFILER_OUTPUT_PATH where profiler outputs written.

    For compatibility, NUX_PROFILER_PATH is also currently being supported, but it will be
    deprecated by FURIOSA_PROFILER_OUTPUT_PATH later.

    Returns:
        The file path of profiler output if specified, or None.
    """
    if os.environ.get("NUX_PROFILER_PATH") is not None:
        warnings.warn("NUX_PROFILER_PATH is deprecated, use FURIOSA_PROFILER_OUTPUT_PATH instead")

    return os.environ.get(
        consts.ENV_FURIOSA_PROFILER_OUTPUT_PATH, os.environ.get("NUX_PROFILER_PATH")
    )



import logging
import os
from pathlib import Path
import random

import numpy as np
import onnxruntime

from furiosa.runtime import session
from tests import test_data

LOGLEVEL = os.environ.get('FURIOSA_LOG_LEVEL', 'INFO').upper()
logging.basicConfig(level=LOGLEVEL)


MNIST_ONNX = test_data("mnist-8.onnx")
MNIST_TFLITE_QUANTIZED = test_data("MNISTnet_uint8_quant_without_softmax.tflite")
NAMED_TENSORS_ONNX = test_data("named_tensors.onnx")


def assert_tensors_equal(expected, result):
    assert np.allclose(expected, result, atol=1.0), "{} was expected, but the result was {}".format(
        expected, result
    )


class SessionTester:
    def __init__(self, model_path):
        self.session = session.create(model=model_path)

    def close(self):
        self.session.close()


class AsyncSessionTester:
    def __init__(self, model_path):
        (self.session, self.queue) = session.create_async(model=Path(model_path))

    def close(self):
        self.queue.close()
        self.session.close()


class PredictionTester:
    def __init__(self, model_path):
        self.sess = onnxruntime.InferenceSession(model_path)

    def _run_nux(self, inputs: np.ndarray):
        pass

    def _run_onnxrt(self, inputs: np.ndarray):
        input_name = self.sess.get_inputs()[0].name
        return self.sess.run(None, {input_name: inputs})

    def assert_equals(self, inputs: np.ndarray):
        tf_results = self._run_onnxrt(inputs)
        nux_results = self._run_nux(inputs)

        assert_tensors_equal(tf_results, nux_results)


class BlockingPredictionTester(PredictionTester):
    def __init__(self, model_path):
        self.nux_sess = session.create(model=model_path)
        super().__init__(model_path)

    def _run_nux(self, inputs: np.ndarray):
        return self.nux_sess.run(inputs)[0].numpy()

    def close(self):
        self.nux_sess.close()


class AsyncPredictionTester(PredictionTester):
    def __init__(self, model_path):
        nux_sess, nux_queue = session.create_async(model=model_path)
        self.nux_sess = nux_sess
        self.nux_queue = nux_queue
        super().__init__(model_path)

    def _run_nux(self, inputs: np.ndarray) -> np.ndarray:
        key = random.randint(0, 100)
        self.nux_sess.submit(inputs, context={'key': key})
        _, outputs = self.nux_queue.recv()
        return outputs[0].numpy()

    def close(self):
        self.nux_queue.close()
        self.nux_sess.close()


def exist_char_dev(dev_path: str) -> bool:
    """
    Return True if a specified device exists, or False
    """
    import stat

    # NPU device is a character device
    return stat.S_ISCHR(os.stat(dev_path).st_mode)


def ensure_test_device() -> bool:
    """
    Return True if a NPU device for unit test is ready, or False.
    """
    device = os.getenv('NPU_DEVNAME')
    return device is not None and exist_char_dev(f"/dev/{device}")



import random
import unittest

import mnist
import numpy as np

from furiosa.runtime import session
from furiosa.runtime.tensor import DataType, TensorArray, numpy_dtype
from tests.test_base import MNIST_ONNX, NAMED_TENSORS_ONNX, SessionTester, assert_tensors_equal


class TestTensor(unittest.TestCase):
    mnist_images = mnist.train_images().reshape((60000, 1, 28, 28)).astype(np.float32)

    @classmethod
    def setUpClass(cls):
        cls.sess = SessionTester(MNIST_ONNX).session

    @classmethod
    def tearDownClass(cls):
        cls.sess.close()

    def test_dtype(self):
        tensor = self.sess.input(0)
        self.assertEqual(np.float32, numpy_dtype(tensor.dtype))

    def test_tensor_desc(self):
        tensor = self.sess.input(0)
        self.assertEqual(4, tensor.ndim)
        self.assertEqual("NCHW", tensor.format)
        self.assertEqual((1, 1, 28, 28), tensor.shape)
        self.assertEqual(DataType.FLOAT32, tensor.dtype)
        self.assertEqual(np.float32, numpy_dtype(tensor))
        self.assertEqual(784, tensor.length)
        self.assertEqual(3136, tensor.size)
        self.assertEqual(
            tensor.__repr__(),
            'TensorDesc(name="Input3", shape=(1, 1, 28, 28), dtype=FLOAT32, format=NCHW, size=3136, len=784)',
        )

    def test_tensor(self):
        sess = self.sess
        self.assertEqual(sess.input_num, 1)
        self.assertEqual(sess.output_num, 1)

        inputs: TensorArray = sess.allocate_inputs()
        inputs2: TensorArray = sess.allocate_inputs()
        self.assertEqual(1, len(inputs))
        self.assertEqual(1, len(inputs[0:1]))
        self.assertEqual(np.float32, numpy_dtype(inputs[0]))

        # Testing TensorArray.__repr__()
        inputs[0] = np.zeros(inputs[0].shape, np.float32)
        self.assertEqual(
            inputs[0].__repr__(), 'Tensor(name="Input3", shape=(1, 1, 28, 28), dtype=FLOAT32)'
        )

        for i in range(0, 10):
            idx = random.randrange(0, 9999, 1)
            ndarray_value = self.mnist_images[idx : idx + 1]

            inputs[0] = ndarray_value
            assert_tensors_equal(ndarray_value, inputs[0].numpy())

            # idempotent
            self.assertEqual(inputs[0], inputs[0])

            # equality of contents
            inputs2[0] = ndarray_value
            self.assertEqual(inputs[0], inputs2[0])


class TestTensorNames(unittest.TestCase):
    def test_tensor_names(self):
        with session.create(NAMED_TENSORS_ONNX) as sess:
            self.assertEqual(sess.input_num, 3)
            self.assertEqual(sess.output_num, 3)
            for idx, expected in enumerate(["input.1", "input.3", "input"]):
                actual_result = sess.input(idx).name
                self.assertEqual(
                    actual_result, expected, f"expected: {expected}, actual: {actual_result}"
                )
            for idx, expected in enumerate(["18_dequantized", "19_dequantized", "15_dequantized"]):
                actual_result = sess.output(idx).name
                self.assertEqual(
                    actual_result, expected, f"expected: {expected}, actual: {actual_result}"
                )

            self.assertEqual(
                sess.input(0).__repr__(),
                "TensorDesc(name=\"input.1\", shape=(1, 3, 8, 8), dtype=FLOAT32, format=NCHW, size=768, len=192)",
            )


if __name__ == '__main__':
    unittest.main()



"""Tensor object and its utilities"""

import ctypes
from ctypes import POINTER, byref, c_uint8, c_uint64, c_void_p
from enum import IntEnum
from typing import Optional, Union

import numpy as np

from ._api import LIBNUX
from ._util import list_to_dict
from .errors import UnsupportedTensorType


class Axis(IntEnum):
    """Axis of Tensor"""

    WIDTH = 0
    HEIGHT = 1
    CHANNEL = 2
    BATCH = 3
    WIDTH_OUTER = 4
    HEIGHT_OUTER = 5
    CHANNEL_OUTER = 6
    BATCH_OUTER = 7
    UNKNOWN = 8

    @classmethod
    def _names(cls):
        return ["W", "H", "C", "N", "Wo", "Ho", "Co", "No", "?"]

    def __repr__(self):
        return self._names()[self]


class DataType(IntEnum):
    """Tensor data type"""

    FLOAT32 = 0
    UINT8 = 1
    INT8 = 2
    INT32 = 3
    INT64 = 4
    BFLOAT16 = 5  # Not supported yet in Numpy

    @classmethod
    def _numpy_dtypes(cls):
        return [np.float32, np.uint8, np.int8, np.int32, np.int64]

    def __repr__(self) -> str:
        return self._name_

    @property
    def numpy_dtype(self):
        """Return the numpy dtype corresponding to this DataType"""
        return self._numpy_dtypes()[self]


class TensorDesc:
    """Tensor description including dimension, shape, and data type"""

    def __init__(self, ref: c_void_p):
        self.ref = ref
        self._as_parameter_ = ref

    @property
    def name(self) -> Optional[str]:
        name_ptr = LIBNUX.nux_tensor_name(self)
        if name_ptr:
            name = ctypes.c_char_p(name_ptr).value.decode("utf-8")
            LIBNUX.nux_string_destroy(name_ptr)
            return name
        else:
            return None

    @property
    def ndim(self) -> int:
        """Number of dimensions"""
        return LIBNUX.nux_tensor_dim_num(self)

    def dim(self, idx: int) -> int:
        """Size of i-th dimension"""
        return LIBNUX.nux_tensor_dim(self, idx)

    @property
    def shape(self) -> tuple:
        """tensor shape"""
        dims = []
        for i in range(self.ndim):
            dims.append(self.dim(i))

        return tuple(dims)

    def axis(self, idx: int) -> Axis:
        """Axis type of i-th dimension (e.g., width, height, channel)"""
        return Axis(LIBNUX.nux_tensor_axis(self, idx))

    @property
    def size(self) -> int:
        """Size in bytes"""
        return LIBNUX.nux_tensor_size(self)

    @property
    def length(self) -> int:
        """Number of all elements across all dimensions"""
        return LIBNUX.nux_tensor_len(self)

    @property
    def format(self) -> str:
        """Tensor memory layout (e.g., NHWC, NCHW)"""
        tensor_format = str()
        for idx in range(self.ndim):
            tensor_format += self.axis(idx).__repr__()

        return tensor_format

    @property
    def dtype(self) -> DataType:
        """Data type of tensor"""
        return DataType(LIBNUX.nux_tensor_dtype(self))

    @property
    def numpy_dtype(self):
        """Return numpy dtype"""
        return self.dtype.numpy_dtype

    def __repr__(self) -> str:
        repr = self.__class__.__name__ + "("
        if self.name:
            repr += f"name=\"{self.name}\", "

        repr += (
            f"shape={self.shape}, dtype={self.dtype.__repr__()}, "
            f"format={self.format}, size={self.size}, len={self.length})"
        )

        return repr


class Tensor:
    """A tensor which contains data and tensor description including shape"""

    def __init__(self, ref: c_void_p, desc: TensorDesc, allocated: bool = False):
        self.ref = ref
        self.desc = desc
        self.allocated = allocated
        self._as_parameter_ = ref

    @property
    def shape(self) -> tuple:
        """Return the tensor shape

        Returns:
            Tensor shape. An example shape is
            ```(1, 28, 28, 1)```.
        """
        return self.desc.shape

    @property
    def dtype(self) -> DataType:
        """Data type of tensor"""
        return self.desc.dtype

    @property
    def numpy_dtype(self):
        """Return numpy dtype"""
        return self.desc.numpy_dtype

    def copy_from(self, data: Union[np.ndarray, np.generic]):
        """Copy the contents of Numpy ndarray to this tensor"""
        if isinstance(data, np.ndarray):
            data = np.ascontiguousarray(data)
        if isinstance(data, (np.ndarray, np.generic)):
            buf_ptr = data.ctypes.data_as(POINTER(c_uint8))
            buf_size_in_bytes = data.nbytes
            LIBNUX.tensor_fill_buffer(self, buf_ptr, buf_size_in_bytes)
        else:
            raise UnsupportedTensorType()

    def numpy(self) -> np.ndarray:
        """Return numpy.ndarray converted from this tensor"""
        itemsize = np.dtype(self.numpy_dtype).itemsize
        arr_size = np.prod(self.shape[:]) * itemsize

        buf_ptr = POINTER(c_uint8)()
        buf_len = c_uint64(0)
        LIBNUX.tensor_get_buffer(self, byref(buf_ptr), byref(buf_len))

        buf_from_mem = ctypes.pythonapi.PyMemoryView_FromMemory
        buf_from_mem.restype = ctypes.py_object
        buf_from_mem.argtypes = (ctypes.c_void_p, ctypes.c_int, ctypes.c_int)
        buffer = buf_from_mem(buf_ptr, arr_size, 0x100)

        arr = np.ndarray(tuple(self.shape[:]), self.numpy_dtype, buffer, order="C")
        return arr.copy()

    def __repr__(self):
        repr = self.__class__.__name__ + "("

        if self.desc.name:
            repr += f'name="{self.desc.name}", '

        repr += f"shape={self.desc.shape}, dtype={self.desc.dtype.__repr__()})"

        return repr

    def __del__(self):
        if self.allocated and self.ref:
            LIBNUX.nux_tensor_destroy(self.ref)
            self.allocated = False

    def __eq__(self, other):
        if isinstance(other, Tensor):
            return np.array_equal(self.numpy(), other.numpy())

        return False


class TensorArray:
    """A list of tensors

    It is used for input and output values of model inferences.
    """

    def __init__(self, ref: c_void_p, descs: [TensorDesc], allocated: bool = False):
        self.ref = ref
        self.descs = descs
        self.should_drop = allocated
        self._as_parameter_ = ref
        self.len = LIBNUX.nux_tensor_array_len(self.ref)

    def is_empty(self) -> bool:
        """True if it has no Tensor"""
        return self.len == 0

    def __len__(self):
        return self.len

    def __getitem__(self, key):
        if isinstance(key, int):
            while key < 0:
                # Index is a negative, so addition will subtract.
                key += len(self)

            if key >= len(self):
                raise IndexError("tensor index (%d) out of range [0, %d)" % (key, len(self)))

            return Tensor(
                LIBNUX.nux_tensor_array_get(self, key), desc=self.descs[key], allocated=False
            )

        if isinstance(key, slice):
            start, stop, step = key.indices(len(self))
            return [self[i] for i in range(start, stop, step)]

        raise TypeError

    def __setitem__(self, key, data):
        if isinstance(key, int):
            while key < 0:
                # Index is a negative, so addition will subtract.
                key += len(self)

            if key >= len(self):
                raise IndexError("tensor index (%d) out of range [0, %d)" % (key, len(self)))

            self[key].copy_from(data)
            return

        raise TypeError

    def numpy(self) -> [np.ndarray]:
        """Convert TensorArray to a list of numpy.ndarray"""
        array = []
        for idx in range(self.len):
            array.append(self[idx].numpy())

        return array

    def __del__(self):
        if self.should_drop and self.ref:
            LIBNUX.nux_tensor_array_destroy(self.ref)
            self.should_drop = False

    def __repr__(self):
        return list_to_dict(self).__repr__()


def numpy_dtype(value):
    """Return numpy dtype from any eligible object of Nux"""
    if isinstance(value, (np.ndarray, np.generic)):
        return value.dtype
    if isinstance(value, Tensor):
        return value.numpy_dtype
    if isinstance(value, TensorDesc):
        return value.numpy_dtype
    if isinstance(value, DataType):
        return value.numpy_dtype

    raise TypeError


def rand(tensor: TensorDesc) -> np.ndarray:
    """Return a new array of given shape and type, filled with random numbers."""
    return np.random.rand(*tensor.shape).astype(tensor.numpy_dtype)


def zeros(tensor: TensorDesc) -> np.ndarray:
    """Return a new array of given shape and type, filled with zeros."""
    return np.zeros(shape=tensor.shape, dtype=tensor.numpy_dtype)



import re
import unittest

from furiosa.runtime.errors import DeviceBusy

versionPattern = r'\d+(=?\.(\d+(=?\.(\d+)*)*)*)*'
regexMatcher = re.compile(versionPattern)


class TestFuriosaError(unittest.TestCase):
    def test_error(self):
        error = DeviceBusy()
        self.assertEqual(error.__repr__(), "NPU device busy (native error code: 23)")
        self.assertEqual(error.__str__(), "NPU device busy (native error code: 23)")
        self.assertEqual(error._native_err, 23)


if __name__ == '__main__':
    unittest.main()



import re
import unittest

from furiosa import runtime
from furiosa.runtime import LIBNUX

versionPattern = r'\d+(=?\.(\d+(=?\.(\d+)*)*)*)*'
regexMatcher = re.compile(versionPattern)


class TestTensor(unittest.TestCase):
    def test_version(self):
        self.assertTrue(len(runtime.__version__.version) > 0)
        self.assertTrue(regexMatcher.match(LIBNUX.version().decode('utf-8')))
        self.assertTrue(len(str(LIBNUX.git_short_hash())) >= 9)
        self.assertEqual(len(str(LIBNUX.build_timestamp())), 22)


if __name__ == '__main__':
    unittest.main()



"""Provide high-level Python APIs to access Furiosa AI's NPUs and its eco-system"""

from furiosa.common.utils import get_sdk_version

from ._api import LIBNUX, runtime_version


def full_version() -> str:
    """Returns a full version string including the native library version"""
    return "Furiosa SDK Runtime {} (libnux {} {} {})".format(
        __version__,
        LIBNUX.version().decode('utf-8'),
        LIBNUX.git_short_hash().decode('utf-8'),
        LIBNUX.build_timestamp().decode('utf-8'),
    )


__version__ = get_sdk_version(__name__)
__full_version__ = full_version()
